import sys
from types import IntType

from qt import *

from kuralib.config import Config
from kuradom import KuraDOM
from kuradbreader import KuraDBReader

def escape(data, entities={}):
    """Escape &, <, and > in a string of data.

    You can escape other strings of data by passing a dictionary as 
    the optional entities parameter.  The keys and values must all be
    strings; each key will be replaced with its corresponding value.
    """
    data = data.replace("&", "&amp;")
    data = data.replace("<", "&lt;")
    data = data.replace(">", "&gt;")
    for chars, entity in entities.items():
        data = data.replace(chars, entity)        
    return data

    
class XMLGenerator:

    def __init__(self, out=None, encoding="iso-8859-1"):
        if out is None:
            import sys
            out = sys.stdout
        self._out = out
        self._encoding = encoding

        self._indent=-1

    def startDocument(self):
        self._out.write('<?xml version="1.0" encoding="%s"?>\n' %
                        self._encoding)

    def startElement(self, name, attrs):
        self._indent+=1
        self._out.write(self._indent * "\t" + '<' + name + "\n")
        for (name, value) in attrs.items():
            self._out.write(self._indent * "\t" + ' %s="%s"' % (name, escape(value))+ "\n")
        self._out.write(self._indent * "\t" + '>'+ "\n")
        
    def endElement(self, name):
        self._out.write(self._indent * "\t" + '</%s>' % name+ "\n")
        self._indent-=1


class KuraXMLWriter(QObject):

  def __init__(self, dom, *args):
    apply(QObject.__init__,(self,)+args)
    self.dom=dom
    self.generator=XMLGenerator(encoding="utf-8")
    
    self.connect(self.dom, PYSIGNAL("sigTextStarted"), self._startText)
    self.connect(self.dom, PYSIGNAL("sigTextEnded"), self._textEnded)
    self.connect(self.dom, PYSIGNAL("sigStreamStarted"), self._startStream)
    self.connect(self.dom, PYSIGNAL("sigStreamEnded"), self._streamEnded)
    self.connect(self.dom, PYSIGNAL("sigElementStarted"), self._startElement)
    self.connect(self.dom, PYSIGNAL("sigElementEnded"), self._elementEnded)
    
    self.connect(self.dom, PYSIGNAL("sigTextTag"), self._startTextTag)
    self.connect(self.dom, PYSIGNAL("sigStreamTag"), self._startStreamTag)
    self.connect(self.dom, PYSIGNAL("sigElementTag"), self._startElementTag)

  def _filterFields(self, fields):
    for k, v in   fields.items():
      if type(v)==IntType:
        fields[k]=str(v)
      if v==None or v=="": 
        del fields[k]
    return fields
      
  def _startText(self, text):
    self.generator.startElement("text", self._filterFields(text.getFields()))
    
  def _textEnded(self, text):
    self.generator.endElement("text")

  def _startStream(self, stream):
    self.generator.startElement("stream", self._filterFields(stream.getFields()))

  def _streamEnded(self, stream):
    self.generator.endElement("stream")
    
  def _startElement(self, element):
    self.generator.startElement("element", self._filterFields(element.getFields()))
      
  def _elementEnded(self, element):
    self.generator.endElement("element")
    
  def _startTextTag(self, tag):
    self.generator.startElement("text_tag", self._filterFields(tag.getFields()))
    self.generator.endElement("text_tag")
    
  def _startStreamTag(self, tag):
    self.generator.startElement("stream_tag", self._filterFields(tag.getFields()))
    self.generator.endElement("stream_tag")
        
  def _startElementTag(self, tag):    
    self.generator.startElement("element_tag", self._filterFields(tag.getFields()))
    self.generator.endElement("element_tag")

  def start(self):
    self.generator.startDocument()
    self.generator.startElement("kuracollection", {})
    self.dom.walkTree()
    self.generator.endElement("kuracollection")
    
    
def main(args):
  dbr=KuraDBReader()
  for arg in args:
    dbr.start(arg)
  xmlsaver=KuraXMLWriter(dbr.getDom())
  xmlsaver.start()
  
if __name__=="__main__":
  main(sys.argv[1:])
